@@ -177,7 +177,7 @@ winUninstallColormap(ColormapPtr pmap)
 
     /* Install the default cmap in place of the cmap to be uninstalled */
     if (pmap->mid != pmap->pScreen->defColormap) {
-        dixLookupResourceByType((pointer) &curpmap, pmap->pScreen->defColormap,
+        dixLookupResourceByType((void *) &curpmap, pmap->pScreen->defColormap,
                                 RT_COLORMAP, NullClient, DixUnknownAccess);
         (*pmap->pScreen->InstallColormap) (curpmap);
     }
